#include "Move.hpp"
#include "Exceptions.hpp"

Move::Move(const MoveType& type, const Coord& from, const Coord& to, const PieceType& promoted = NOT_A_TYPE)
    : type(type), from(from), to(to), promoted(promoted)
{
    // none
}

Move Move::none() {
    return Move(MoveType::NONE, Coord(1,1), Coord(1,1));
}

Move Move::silent(const Coord& from, const Coord& to) {
    return Move(MoveType::SILENT, from, to);
}

Move Move::pawn_double(const Coord& from, const Coord& to) {
	return Move(MoveType::PAWN_DOUBLE, from, to);
}

Move Move::capture(const Coord& from, const Coord& to) {
    return Move(MoveType::CAPTURE, from, to);
}

Move Move::enpassant(const Coord& from, const Coord& to) {
    return Move(MoveType::ENPASSANT, from, to);
}

Move Move::rooking(const Coord& from, const Coord& to) {
    return Move(MoveType::ROOKING, from, to);
}

Move Move::promotion(const Coord& from, const Coord& to, const PieceType& promoted) {
    return Move(MoveType::PROMOTION, from, to, promoted);
}

Move Move::from_fen(std::string fen, const Board& board) {
    const Coord from = Coord::from_fen( fen.substr(0,2) );
    const Coord to = Coord::from_fen( fen.substr(2,2) );
	const auto& from_piece = board.get_piece_at(from);

    // rooking
    if (from_piece.type == KING) {
		auto file_delta = (from.file > to.file) ? from.file - to.file : to.file - from.file;
		if (file_delta == 2) {
			if (to.file == 3) return Move::rooking(from, to);
			if (to.file == 7) return Move::rooking(from, to);
		}
    }

    // enpassant
    if (to.rank == 3 || to.rank == 6) {
		if (from_piece.type == PAWN && board.enpassant_posibility.coord && to == board.enpassant_posibility.coord.get()) {
            const Coord before_to(to.file, (to.rank>=5)? 5:4);
            if (board.has_piece_at(before_to) && board.get_piece_at(before_to).type == PAWN) {
                if (board.get_piece_at(before_to).color != board.get_piece_at(from).color) {
                    return Move::enpassant(from, to);
                }
            }
        }
    }

    // promotion
	if (from_piece.type == PAWN && (to.rank == 1 || to.rank == 8)) {
        const Piece promoted = Piece::from_fen(fen.substr(4,1));
        return Move::promotion(from, to, promoted.type);
    }

    // capture
    if (board.has_piece_at(to)) return Move::capture(from, to);

	// pawn double
	if (from_piece.type == PAWN) {
		UNumber start_rank = (from_piece.color == WHITE) ? 2 : 7;
		UNumber double_rank = (from_piece.color == WHITE) ? 4 : 5;
		if ((from.rank == start_rank) && (to.rank == double_rank)) {
			return Move::pawn_double(from, to);
		}
	}

    // silent
    return Move::silent(from, to);
}

std::string Move::to_fen() const {

    switch((*this).type) {
        case MoveType::NONE:
		case MoveType::SILENT:
		case MoveType::PAWN_DOUBLE:
		case MoveType::CAPTURE:
		case MoveType::ENPASSANT:
        case MoveType::ROOKING:
    return from.to_fen() + to.to_fen();
        case MoveType::PROMOTION:
            return from.to_fen() + to.to_fen() + Piece(WHITE, promoted).to_fen();
    }

    throw CustomException("Some move is not to_fenable.");
}

RestoreData make_silent(const Move& move, Board& board) {
	auto piece = board.get_piece_at(move.from);
	board.move_piece(move.from, move.to);

	RestoreData result(board.enpassant_posibility, board.rooking_posibility);
	board.enpassant_posibility = EnpassantPosibility::none();

	result.add_action(Action::move(move.to, move.from));

	return result;
}

RestoreData make_pawn_double(const Move& move, Board& board) {
	board.move_piece(move.from, move.to);

	RestoreData result(board.enpassant_posibility, board.rooking_posibility);
	board.enpassant_posibility = EnpassantPosibility::none();

	const UNumber ep_rank = (move.to.rank == 4) ? 3 : 6;
	const Coord ep(move.from.file, ep_rank);
	board.enpassant_posibility = EnpassantPosibility(ep);

	result.add_action(Action::move(move.to, move.from));

	return result;
}

RestoreData make_capture(const Move& move, Board& board) {
    auto piece = board.get_piece_at(move.from);
    auto captured = board.get_piece_at(move.to);
    board.remove_piece(move.to);
    board.move_piece(move.from, move.to);

	RestoreData result(board.enpassant_posibility, board.rooking_posibility);

    result.add_action(Action::move(move.to, move.from));
    result.add_action(Action::set(captured, move.to));
    board.enpassant_posibility = EnpassantPosibility::none();

    return result;
}

RestoreData make_enpassant(const Move& move, Board& board) {
    auto piece = board.get_piece_at(move.from);
    auto enpassant_coord = board.enpassant_posibility.coord.get();
    auto capture_coord = Coord(enpassant_coord.file, enpassant_coord.rank == 3? 4: 5);
    auto captured = board.get_piece_at(capture_coord);
    board.remove_piece(capture_coord);
    board.move_piece(move.from, move.to);

    RestoreData result(board.enpassant_posibility, board.rooking_posibility);

    result.add_action(Action::move(move.to, move.from));
    result.add_action(Action::set(captured, capture_coord));
    board.enpassant_posibility = EnpassantPosibility::none();

    return result;
}

RestoreData make_rooking(const Move& move, Board& board) {
    UNumber rook_rank = move.to.rank;
    UNumber rook_from_file = 0;
    UNumber rook_to_file = 0;
    if (move.to.file == 3) { // long
        rook_from_file = 1;
        rook_to_file = 4;
    }
    if (move.to.file == 7) { // short
        rook_from_file = 8;
        rook_to_file = 6;
    }
	auto rook_from = Coord(rook_from_file, rook_rank);
	auto rook_to = Coord(rook_to_file, rook_rank);
	auto rook = board.get_piece_at(rook_from);

    board.move_piece(move.from, move.to);
	board.move_piece(rook_from, rook_to);

    RestoreData result(board.enpassant_posibility, board.rooking_posibility);

    result.add_action(Action::move(move.to, move.from));
	result.add_action(Action::set(rook, rook_from));
	result.add_action(Action::remove(rook_to));
    board.enpassant_posibility = EnpassantPosibility::none();

	return result;
}

RestoreData make_promotion(const Move& move, Board& board) {
    auto color = board.get_piece_at(move.from).color;

    RestoreData result(board.enpassant_posibility, board.rooking_posibility);

    board.remove_piece(move.from);
    result.add_action(Action::set(Piece(color, PAWN), move.from));
    result.add_action(Action::remove(move.to));

    if (board.has_piece_at(move.to)) {
        auto removed = board.get_piece_at(move.to);
        board.remove_piece(move.to);
        result.add_action(Action::set(removed, move.to));
    }

    board.add_piece(Piece(color, move.promoted), move.to);

    board.enpassant_posibility = EnpassantPosibility::none();

    return result;
}

std::pair<Coord, Coord> Move::pair() const {
    return std::pair<Coord, Coord>(from, to);
}

bool Move::is_silent() const {
    if ((*this).type == MoveType::SILENT) return true;
    if ((*this).type == MoveType::PAWN_DOUBLE) return true;

    return false;
}

bool Move::is_capture() const {
    if ((*this).type == MoveType::CAPTURE) return true;
    if ((*this).type == MoveType::ENPASSANT) return true;

    return false;
}

bool Move::is_enpassant() const {
    return ((*this).type == MoveType::ENPASSANT);
}

bool Move::is_promotion() const {
    return ((*this).type == MoveType::PROMOTION);
}

bool Move::is_rooking() const {
    return ((*this).type == MoveType::ROOKING);
}


RestoreData Move::make(Board& board) const {
	RestoreData result(board.enpassant_posibility, board.rooking_posibility);

	board.turn_color = board.turn_color.inverse();

	auto new_rp = RookingPosibility::all();
	if (board.has_piece_at(to)) { // this is a capture
		auto captured = board.get_piece_at(to);
		if (captured.type == ROOK) {
			if (captured.color == WHITE && to == Coord::from_fen("h1")) {
				new_rp.white_king = false;
			}
			else if (captured.color == WHITE && to == Coord::from_fen("a1")) {
				new_rp.white_queen = false;
			}
			else if (captured.color == BLACK && to == Coord::from_fen("h8")) {
				new_rp.black_king = false;
			}
			else if (captured.color == BLACK && to == Coord::from_fen("a8")) {
				new_rp.black_queen = false;
			}
		}
	}

	if ((*this).type == MoveType::SILENT) result = make_silent(*this, board);
	if ((*this).type == MoveType::PAWN_DOUBLE) result = make_pawn_double(*this, board);
    if ((*this).type == MoveType::CAPTURE) result = make_capture(*this, board);
    if ((*this).type == MoveType::ENPASSANT) result = make_enpassant(*this, board);
    if ((*this).type == MoveType::ROOKING) result = make_rooking(*this, board);
	if ((*this).type == MoveType::PROMOTION) result = make_promotion(*this, board);

	if (new_rp.white_king == false) board.rooking_posibility.white_king = false;
	if (new_rp.white_queen == false) board.rooking_posibility.white_queen = false;
	if (new_rp.black_king == false) board.rooking_posibility.black_king = false;
	if (new_rp.black_queen == false) board.rooking_posibility.black_queen = false;

	if (board.has_piece_at(to) && board.get_piece_at(to).type == KING) {
		if (board.get_piece_at(to).color == WHITE) {
			board.rooking_posibility.white_king = false;
			board.rooking_posibility.white_queen = false;
		}
		else {
			board.rooking_posibility.black_king = false;
			board.rooking_posibility.black_queen = false;
		}
	}

	if (board.has_piece_at(to) && board.get_piece_at(to).type == ROOK) {
		if (from == Coord::from_fen("a1")) {
			board.rooking_posibility.white_queen = false;
		}
		else if (from == Coord::from_fen("h1")) {
			board.rooking_posibility.white_king = false;
		}
		else if (from == Coord::from_fen("a8")) {
			board.rooking_posibility.black_queen = false;
		}
		else if (from == Coord::from_fen("h8")) {
			board.rooking_posibility.black_king = false;
		}
	}

    result.last_moved = board.last_moved;

    return result;
}

bool Move::operator==(const Move& rhs) const {
    return (*this).type == rhs.type && (*this).from == rhs.from && (*this).to == rhs.to && (*this).promoted == rhs.promoted;
}

bool Move::operator!=(const Move& rhs) const {
    return (*this).type != rhs.type || (*this).from != rhs.from || (*this).to != rhs.to || (*this).promoted != rhs.promoted;
}
