#include <algorithm>
#include <iostream>
#include <vector>
#include <chrono>
#include <thread>
#include <cstdio>
#include <cmath>

#include "Search.hpp"
#include "Generator.hpp"
#include "Exceptions.hpp"
#include "Checker.hpp"

// for clang and gcc compilers
#define _GLIBCXX_USE_NANOSLEEP 1

// TODO: to move
bool inlegal(Board& board, const Move& move) {
    auto data = move.make(board);
    auto result = Checker::is_check(board, board.turn_color.inverse());
    data.restore(board);
    return result;
};

// TODO: to sort (or hash)
void sort_by_hash(Moves& moves, const Move& hash_move) {
    for (auto it = moves.begin(); it != moves.end(); ++it) {
        if (it->second == hash_move) {
            auto saved = *it;
            moves.erase(it);
            moves.push_front(saved);
            break;
        }
    }
}

// TODO: to sort (or history)
void sort_by_history(Moves& moves) {
    using SM = std::pair<Score, Move>;
    auto history_pred = [](SM sm1, SM sm2) {
        auto move1 = sm1.second;
        auto move2 = sm2.second;

        if (move2.is_capture()) return false; // dont sort captures

        auto score1 = Search::Manager::history.get(move1);
        auto score2 = Search::Manager::history.get(move2);

        return score1 > score2;
    };

    moves.sort(history_pred);
}

void sort_by_last_moved(const Board& board, Moves& moves) {
    for (auto it = moves.begin(); it != moves.end(); ++it) {
        if (it->second.to == board.last_moved) {
            auto saved = *it;
            moves.erase(it);
            moves.push_front(saved);
            break;
        }
    }
}

using Search_Results = std::pair<Score, std::list<Move>>;

UNumber max_quies_depth = 0;
Search_Results quies(Board& board, Score alpha, Score beta) {
	auto null = Search_Results(-Evaluation::max_score, std::list<Move>());

	++Search::Manager::nodes;
	if (Search::Manager::stoped) return null;

	if (Search::Manager::last_positions.has(board))
		return Search_Results(0, std::list<Move>());

	Score score = Evaluation::total(board);

	alpha = std::max(alpha, score);

	std::list<Move> pv;
    if (alpha >= beta) return Search_Results(alpha, pv);

	Search::Manager::last_positions.add(board, Hash::Record());

    auto hash_move = Move::none();
    if (Search::Manager::table.has(board)) {
        auto record = Search::Manager::table.get(board);
        pv = record.pv;
        hash_move = pv.front();
    }

    auto moves = Generator::for_board(board);

    sort_by_last_moved(board, moves);

    if (hash_move != Move::none()) {
        sort_by_hash(moves, hash_move);
    }

    bool has_move = false;
    auto result_score = score;
	for (auto scored_move : moves) {
        auto move = scored_move.second;

        auto restore_data = move.make(board);

        auto inlegal = Checker::is_check(board, board.turn_color.inverse());
        if (inlegal) {
            restore_data.restore(board);
            continue;
        }

        has_move = true;

		if (!move.is_capture() && !move.is_promotion() && !Checker::is_check(board)) {
			restore_data.restore(board);
			continue;
        }


        static UNumber quies_depth = 1;
        ++quies_depth;
        max_quies_depth = std::max(quies_depth, max_quies_depth);
        auto search_results = quies(board, -beta, -alpha);
		--quies_depth;
        score = -search_results.first;

		restore_data.restore(board);

        if (score > result_score) result_score = score;
		if (score > alpha) {
			alpha = score;
			pv = search_results.second;
			pv.push_front(move);
		}

		if (score >= beta) {
			Search::Manager::last_positions.remove(board);
			return Search_Results(score, pv);
		}
	}

    if (!has_move) {
        if (Checker::is_check(board))
            result_score = -Evaluation::mate();
        else
            result_score = 0;
    }

	Search::Manager::last_positions.remove(board);

	return Search_Results(result_score, pv);
}

UNumber max_negascout_depth = 0;
Search_Results negascout(Board& board, Score alpha, Score beta, UNumber depth, bool extended = false) {
	auto ply = (Search::Manager::current_depth - depth);
    auto null = Search_Results(-Evaluation::max_score, std::list<Move>());
    ++Search::Manager::nodes;
    if (Search::Manager::stoped) return null;

	std::list<Move> pv;

	if (Search::Manager::last_positions.has(board)) {
		return Search_Results(0, pv);
	}

    auto old_alpha = alpha;

    auto hash_move = Move::none();
    if (Search::Manager::table.has(board)) {
        auto record = Search::Manager::table.get(board);
        pv = record.pv;
        hash_move = pv.front();

        if (record.depth >= depth && ply > 0) {
            switch (record.nature) {
            case Hash::Record::SCORE_REAL:
                return Search_Results(record.score, pv);
            case Hash::Record::SCORE_MIN:
                if (record.score >= beta) return Search_Results(record.score, pv);
                alpha = std::max(alpha, record.score);
            case Hash::Record::SCORE_MAX:
                if (record.score <= alpha) return Search_Results(record.score, pv);
            }
        }
    }

    if (depth <= 0) return quies(board, alpha, beta);

	Search::Manager::last_positions.add(board, Hash::Record());

    auto moves = Generator::for_board(board);

    sort_by_history(moves);
    
    sort_by_last_moved(board, moves);

    if (hash_move != Move::none()) {
        sort_by_hash(moves, hash_move);
    }
    
    auto result_score = -Evaluation::max_score;
    bool first_move = true;

    bool has_move = false;
	for (auto scored_move : moves) {
		if (Search::Manager::stoped) return null;

		auto move = scored_move.second;

        auto restore_data = move.make(board);

        auto inlegal = Checker::is_check(board, board.turn_color.inverse());
        if (inlegal) {
            restore_data.restore(board);
            continue;
        }

        has_move = true;

        auto R = Checker::is_check(board)? 0: 1;
        auto next_extended = (R == 0) || extended;

        static UNumber negascout_depth = 1;
        ++negascout_depth;
        max_negascout_depth = std::max(negascout_depth, max_negascout_depth);
        Search_Results search_results;
        if (first_move) {
            search_results = negascout(board, -beta, -alpha, depth - R, next_extended);
            first_move = false;
        }
        else {
            search_results = negascout(board, -alpha - 1, -alpha, depth - R, next_extended);
            auto temp_score = -search_results.first;
            if (temp_score > alpha && temp_score < beta)
                search_results = negascout(board, -beta, -temp_score, depth - R, next_extended);
        }
        --negascout_depth;

        auto score = -search_results.first;

        restore_data.restore(board);

        if (score > result_score) result_score = score;
        if (score > alpha) {
            Search::Manager::history.renew(move, alpha, beta, score);

            alpha = score;
            pv = search_results.second;
            pv.push_front(move);

            auto nature = (alpha < beta)? Hash::Record::SCORE_REAL: Hash::Record::SCORE_MIN;
            Hash::Record record = { depth, alpha, nature, pv };
            Search::Manager::table.renew(board, record);
        }

        // alpha/beta pruning
		if (alpha >= beta) {
			Search::Manager::last_positions.remove(board);
			return Search_Results(alpha, pv);
		}
    }

	Search::Manager::last_positions.remove(board);

	if (!has_move) {
		if (Checker::is_check(board))
			result_score = -Evaluation::mate() + ply;
		else
			result_score = 0;

        auto nature = Hash::Record::SCORE_REAL;
        Hash::Record record = { depth, result_score, nature, pv };
        Search::Manager::table.renew(board, record);
	} else if (alpha == old_alpha) {
        auto nature = Hash::Record::SCORE_MAX;
        Hash::Record record = { depth, result_score, nature, pv };
        Search::Manager::table.renew(board, record);
    }

    return Search_Results(result_score, pv);
}

Search_Results negascout_root(Board& board, UNumber depth) {
	Search::Manager::last_positions = Hash::Table::empty();
	Search::Manager::current_depth = depth;
    max_quies_depth = 0;
	return negascout(board, -Evaluation::max_score, Evaluation::max_score, depth);
}

Search_Results aspiration_search(Board& board, Score window, UNumber depth) {
    auto static_score = Evaluation::total(board);
    auto alpha = static_score - window;
    auto beta = static_score + window;

    auto search_results = negascout(board, alpha, beta, depth);
    auto score = search_results.first;

    if (score <= alpha) {
        alpha = -Evaluation::max_score;
        search_results = negascout(board, alpha, beta, depth);
    } else if (score >= beta) {
        beta = Evaluation::max_score;
        search_results = negascout(board, alpha, beta, depth);
    }

    return search_results;
}

Move Search::best_move(Board& board, UNumber depth) {
    auto best_pair = negascout_root(board, depth);
    if (best_pair.second.size() > 0)
        return best_pair.second.front();
    else
        return Move::none();
}

UNumber Search::Manager::depth = -1; // as infinity, not bug
UNumber Search::Manager::current_depth; // not initialized
UNumber Search::Manager::movetime = -1; // same
UNumber Search::Manager::nodes = 0;
Board __e = Board::empty();
Board& Search::Manager::board = __e;
Hash::Table __he = Hash::Table::empty();
Hash::Table& Search::Manager::table = __he;
Hash::Table Search::Manager::last_positions;
History::Table Search::Manager::history;
bool Search::Manager::stoped = true;
Move Search::Manager::best_move = Move::none();

void Search::Manager::reset() {
    join_all();
	if (!stoped) throw CustomException("Can't run another search.");
	best_move = Move::none();
	depth = -1; // not error
    movetime = -1; // same
    nodes = 0;
}

std::vector<std::thread> works;

void stop_in(UNumber ms) {
    const UNumber accuracy = 100;

	while (ms != 0 && !Search::Manager::stoped) {
        std::chrono::milliseconds interval(accuracy);
        std::this_thread::sleep_for(interval);
        ms = (ms <= accuracy)? 0: ms - accuracy;
	}

    Search::Manager::stop();
    return;
}

void Search::Manager::start() {
	if (!stoped) throw CustomException("I can't search 2 best moves at time.");
	stoped = false;
	join_all();

    if (movetime != -1) works.push_back(std::thread(stop_in, movetime));

	works.push_back(std::thread(inc_search));
}

void Search::Manager::clear() {
	stop();
    join_all();
}

void Search::Manager::join_all() {
	for (auto& work : works) work.join();
	works.clear();
}

void Search::Manager::stop() {
	if (stoped) return;
	stoped = true;
	std::cout << "bestmove " << best_move.to_fen() << "\n" << std::flush;
}

void print_info(const std::list<Move>& pv, const Score& score_i, const double time) {
    std::string score_prefix = Evaluation::is_mate(score_i)? "mate": "cp";
    SNumber score_postfix = Evaluation::is_mate(score_i)? Evaluation::get_mate_half_moves(score_i): score_i / 10;
    UNumber nps = Search::Manager::nodes / time;
    std::string pv_string;
    for (auto move: pv) pv_string += move.to_fen() + " ";

    std::cout << "info "
        << "depth " << Search::Manager::current_depth << " "
        << "seldepth " << max_negascout_depth + max_quies_depth << " "
        << "score " << score_prefix << " " << score_postfix << " "
        << "nps " << nps << " "
        << "nodes " << Search::Manager::nodes << " "
        << "pv " << pv_string
        << "\n" << std::flush;
}

void Search::Manager::inc_search() {
	for (UNumber d = 1; d <= depth; ++d) {
        using namespace std::chrono;

        Search::Manager::current_depth = d;

        auto t1 = high_resolution_clock::now();
        auto search_results = negascout_root(board, d);
        auto t2 = high_resolution_clock::now();
        double time = duration_cast<duration<double>>(t2 - t1).count();

        auto pv = search_results.second;
        Search::Manager::best_move = (pv.size() > 0)? pv.front(): Move::none();
        auto score = search_results.first;

		if (stoped) return;
        print_info(pv, score, time);

		// Todo: if debug mode
        std::cout << "info string Branching Factor is " << std::pow(nodes, 1.0/d) << "\n";
	}

	stop();
}
