/* TODO:
- RookingPossibility refactoring
- 1 and 8 as constants (boardSize, maxCoord or something else) in Coord or Board (or Chess.hpp)
- ^ all chess coordnate constants as pawn_start_rank(BLACK) and similar (Chess.hpp)
- make move and data restore must affect on moves_count and unused_moves_count
- managers refactoring
- Move refactoring
- Moves.hpp, Moves.cpp
- SortMoves as module
*/

#include <map>
#include <string>
#include <thread>
#include <vector>
#include <iterator>

#include "Piece.hpp"
#include "Coord.hpp"
#include "Board.hpp"
#include "Move.hpp"
#include "Generator.hpp"
#include "Checker.hpp"
#include "Exceptions.hpp"
#include "Evaluation.hpp"
#include "Search.hpp"

#include "cst.hpp"

void testPiece() {
    CHECK(Piece::from_fen("P") == Piece(WHITE, PAWN));
    CHECK(Piece::from_fen("k") == Piece(BLACK, KING));

    CHECK(Piece(BLACK, KING).to_fen() == "k");
    CHECK(Piece(WHITE, QUEEN).to_fen() == "Q");

    CHECK(Piece(BLACK, BISHOP) == Piece(BLACK, BISHOP));
    CHECK(Piece(BLACK, BISHOP) != Piece(WHITE, BISHOP));
    CHECK(Piece(BLACK, BISHOP) != Piece(BLACK, KNIGHT));
    CHECK(Piece(BLACK, BISHOP) != Piece(WHITE, KING));

    EXCEPTION(Piece::from_fen("g"));
    EXCEPTION(Piece().to_fen());
}

void testCoord() {
    CHECK(Coord(3, 2) == Coord(3, 2));
    CHECK(!(Coord(3, 4) == Coord(3, 2)));
    CHECK(!(Coord(3, 4) == Coord(1, 4)));

    CHECK(Coord::from_fen("a3") == Coord(1, 3));
    CHECK(Coord::from_fen("h8") == Coord(8, 8));

    CHECK(Coord::from_fen("d4").to_fen() == "d4");

    CHECK(Coord(2, 2) < Coord(3, 2));
    CHECK(Coord(4, 5) < Coord(1, 6));

    EXCEPTION(Coord(0,1));
    EXCEPTION(Coord(1,0));
    EXCEPTION(Coord(9,2));
    EXCEPTION(Coord(2,9));
    EXCEPTION(Coord::from_fen("4#"));

    CHECK(Coord::is_valid(1,3));
    CHECK(!Coord::is_valid(5,12));
}

void testColor() {
    CHECK(WHITE.to_fen() == "w");
    CHECK(BLACK.to_fen() == "b");

    CHECK(Color::from_fen("w") == WHITE);
    CHECK(Color::from_fen("b") == BLACK);

    CHECK(WHITE.inverse() == BLACK);
    CHECK(BLACK.inverse() == WHITE);

    EXCEPTION(Color::from_fen("d"));
}

void testEnpassantPosibility() {
    CHECK(EnpassantPosibility::none() == EnpassantPosibility::none());
    CHECK(EnpassantPosibility(Coord(1, 2)) == EnpassantPosibility(Coord(1, 2)));

    CHECK(EnpassantPosibility::from_fen("-") == EnpassantPosibility::none());
    CHECK(EnpassantPosibility::from_fen("d4") == EnpassantPosibility(Coord::from_fen("d4")));

    CHECK(EnpassantPosibility::none().to_fen() == "-");
    CHECK(EnpassantPosibility(Coord(1, 1)).to_fen() == "a1");

    CHECK(EnpassantPosibility(Coord(1, 2)) != EnpassantPosibility::none());
    CHECK(EnpassantPosibility(Coord(1, 2)) != EnpassantPosibility(Coord(4, 6)));

    EXCEPTION(EnpassantPosibility::from_fen("4D"));
}

void testRookingPosibility() {
    CHECK(RookingPosibility(true, true, true, true) == RookingPosibility(true, true, true, true));

    CHECK(RookingPosibility::from_fen("-") == RookingPosibility(false, false, false, false));
    CHECK(RookingPosibility::from_fen("Kq") == RookingPosibility(true, false, false, true));

    CHECK(RookingPosibility::from_fen("-") == RookingPosibility::none());
    CHECK(RookingPosibility::from_fen("KQkq") == RookingPosibility::all());

    CHECK(RookingPosibility::from_fen("Kq").to_fen() == "Kq");
    CHECK(RookingPosibility::from_fen("Qk").to_fen() == "Qk");

    CHECK(RookingPosibility(true, true, true, true) != RookingPosibility(false, true, true, true));
    CHECK(RookingPosibility(true, true, true, true) != RookingPosibility(true, false, true, true));
    CHECK(RookingPosibility(true, true, true, true) != RookingPosibility(true, true, false, true));
    CHECK(RookingPosibility(true, true, true, true) != RookingPosibility(true, true, true, false));

    EXCEPTION(RookingPosibility::from_fen("GdsS"));
}

void testBoard() {
    CHECK(Board() == Board());

    std::string initial_fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";
    std::string empty_fen = "8/8/8/8/8/8/8/8 w - - 0 1";
    CHECK(Board::from_fen(initial_fen).to_fen() == initial_fen);
    CHECK(Board::from_fen(empty_fen).to_fen() == empty_fen);

    CHECK(Board::initial() == Board::from_fen(initial_fen));

    CHECK(Board::from_fen(initial_fen).turn_color == WHITE);
    CHECK(Board::from_fen(empty_fen).turn_color == WHITE);

    CHECK(Board::from_fen(initial_fen).rooking_posibility == RookingPosibility::all());
    CHECK(Board::from_fen(empty_fen).rooking_posibility == RookingPosibility::none());

    CHECK(Board::from_fen(initial_fen).enpassant_posibility == EnpassantPosibility::none());
    CHECK(Board::from_fen(empty_fen).enpassant_posibility == EnpassantPosibility::none());

    CHECK(Board::from_fen(initial_fen).unused_moves_count == 0);
    CHECK(Board::from_fen(empty_fen).unused_moves_count == 0);

    CHECK(Board::from_fen(initial_fen).moves_count == 1);
    CHECK(Board::from_fen(empty_fen).moves_count == 1);

    Board b = Board::from_fen(empty_fen);
    b.add_piece(Piece(WHITE, KING), Coord(1, 1));
    b.add_piece(Piece(BLACK, KING), Coord(8, 8));
    CHECK(b.to_fen() == "7k/8/8/8/8/8/8/K7 w - - 0 1");
    CHECK(!b.is_empty());
    CHECK(b.pieces_count() == 2);

    b.remove_piece(Coord(1, 1));
    b.remove_piece(Coord(8, 8));
    CHECK(b.to_fen() == empty_fen);
    CHECK(b.is_empty());
    CHECK(b.pieces_count() == 0);

    CHECK(b == Board::from_fen(empty_fen));
    CHECK(b != Board::initial());

    EXCEPTION(Board::from_fen("lalala"));
    EXCEPTION(b.add_piece(Piece(BLACK, BISHOP), Coord(12,3)));
    EXCEPTION(b.remove_piece(Coord(1,2)));

    b.add_piece(Piece(BLACK, BISHOP), Coord(1,2));
    EXCEPTION(b.add_piece(Piece(BLACK, BISHOP), Coord(1,2)));

    CHECK(b.has_piece_at(Coord(1,2)));
    CHECK(!b.has_piece_at(Coord(3,2)));
}

void testMove() {
    Move none = Move::none();
    Coord from(1,2);
    Coord to(2,2);
    Move silent = Move::silent(from, to);
    Move capture = Move::capture(from, to);
    Move enpassant = Move::enpassant(from, to);
    Move rooking = Move::rooking(from, to);
    Move promotion = Move::promotion(from, to, QUEEN);

    CHECK(none == Move::none());
    CHECK(silent == Move::silent(from, to));
    CHECK(capture == Move::capture(from, to));
    CHECK(enpassant == Move::enpassant(from, to));
    CHECK(rooking == Move::rooking(from, to));
    CHECK(promotion == Move::promotion(from, to, QUEEN));

    CHECK(silent != none);
    CHECK(silent != capture);
    CHECK(silent != enpassant);
    CHECK(silent != promotion);
    CHECK(silent != rooking);

    CHECK(none.to_fen() == "a1a1");
    CHECK(silent.to_fen() == "a2b2");
    CHECK(capture.to_fen() == "a2b2");
    CHECK(enpassant.to_fen() == "a2b2");
    CHECK(promotion.to_fen() == "a2b2Q");
    CHECK(Move::rooking(Coord(5, 1), Coord(3, 1)).to_fen() == "e1c1");
    CHECK(Move::rooking(Coord(5, 1), Coord(7, 1)).to_fen() == "e1g1");

    Board board = Board::initial();
    CHECK(Move::from_fen("a2a4", board).to_fen() == "a2a4");
    CHECK(Move::from_fen("c2a7", board).to_fen() == "c2a7");
    CHECK(Move::from_fen("c2a8N", board).to_fen() == "c2a8N");
    CHECK(Move::from_fen("e1c1", board).to_fen() == "e1c1");
    CHECK(Move::from_fen("e1g1", board).to_fen() == "e1g1");

    board.remove_piece(Coord::from_fen("f2"));
    board.add_piece(Piece(WHITE, PAWN), Coord::from_fen("f4"));
    board.add_piece(Piece(BLACK, PAWN), Coord::from_fen("e4"));
}

void testGenerator() {
    // Pawns
        Board board = Board::initial();
        Coord e2 = Coord::from_fen("e2");
        Coord e7 = Coord::from_fen("e7");

        CHECK(board.get_piece_at(e2) == Piece(WHITE, PAWN));

        // start
		auto moves = Generator::for_piece(board, e2);
		CHECK(moves.size() == 2);
        CHECK(Generator::for_piece(board, e7).size() == 2);

        // not start
        board.add_piece(Piece(WHITE, PAWN), Coord(1,3));
        auto notstartmoves = Generator::for_piece(board, Coord(1,3));

        CHECK(notstartmoves.size() == 1);

        // captures
        board.add_piece(Piece(WHITE, PAWN), Coord(2,4));
        auto noncaptures = Generator::for_piece(board, Coord(1,3));
        CHECK(noncaptures.size() == 1);

        board.remove_piece(Coord(2,4));
        board.add_piece(Piece(BLACK, PAWN), Coord(2,4));
        auto captures = Generator::for_piece(board, Coord(1,3));
        CHECK(captures.size() == 2);

        // promitions
        Board empty = Board::empty();
        empty.add_piece(Piece(WHITE, PAWN), Coord(1,7));
        auto promitions = Generator::for_piece(empty, Coord(1,7));
        CHECK(promitions.size() == 4);

    // Knights
        board = Board::initial();
        auto b1 = Coord::from_fen("b1");
        auto g1 = Coord::from_fen("g1");
        auto b8 = Coord::from_fen("b8");
        auto g8 = Coord::from_fen("g8");

        CHECK(Generator::for_piece(board, b1).size() == 2);
        CHECK(Generator::for_piece(board, g1).size() == 2);
        CHECK(Generator::for_piece(board, b8).size() == 2);
        CHECK(Generator::for_piece(board, g8).size() == 2);

    // Kings
        empty = Board::empty();
        auto e5 = Coord::from_fen("e5");
        empty.add_piece(Piece(WHITE, KING), e5);

        CHECK(Generator::for_piece(empty, e5).size() == 8);

        empty.add_piece(Piece(WHITE, PAWN), Coord::from_fen("e4"));
        CHECK(Generator::for_piece(empty, e5).size() == 7);

        empty.add_piece(Piece(BLACK, PAWN), Coord::from_fen("f5"));
        CHECK(Generator::for_piece(empty, e5).size() == 7);

        auto initial = Board::initial();
        initial.remove_piece(Coord::from_fen("b1"));
        initial.remove_piece(Coord::from_fen("c1"));
        initial.remove_piece(Coord::from_fen("d1"));
        initial.remove_piece(Coord::from_fen("f1"));
        initial.remove_piece(Coord::from_fen("g1"));
        initial.remove_piece(Coord::from_fen("b8"));
        initial.remove_piece(Coord::from_fen("c8"));
        initial.remove_piece(Coord::from_fen("d8"));
        initial.remove_piece(Coord::from_fen("f8"));
        initial.remove_piece(Coord::from_fen("g8"));

        CHECK(Generator::for_piece(initial, Coord::from_fen("e1")).size() == 4);
        CHECK(Generator::for_piece(initial, Coord::from_fen("e8")).size() == 4);

        CHECK(initial.line_is_empty(Coord::from_fen("e1"), Coord::from_fen("g1")));
        CHECK(initial.line_is_empty(Coord::from_fen("e1"), Coord::from_fen("c1")));
        CHECK(initial.line_is_empty(Coord::from_fen("e8"), Coord::from_fen("g8")));
        CHECK(initial.line_is_empty(Coord::from_fen("e8"), Coord::from_fen("c8")));
        CHECK(initial.line_is_empty(Coord::from_fen("g1"), Coord::from_fen("e1")));
        CHECK(initial.line_is_empty(Coord::from_fen("c1"), Coord::from_fen("e1")));
        CHECK(initial.line_is_empty(Coord::from_fen("g8"), Coord::from_fen("e8")));
        CHECK(initial.line_is_empty(Coord::from_fen("c8"), Coord::from_fen("e8")));

    // Rooks
        empty = Board::empty();

        for (UNumber file = 1; file <= 8; ++file)
            for (UNumber rank = 1; rank <= 8; ++rank)
            {
                auto cell = Coord(file, rank);
                empty.add_piece(Piece(WHITE, ROOK), cell);
                CHECK(Generator::for_piece(empty, cell).size() == 14);
                empty.remove_piece(cell);
            }

    // Bishops
        empty.add_piece(Piece(WHITE, BISHOP), e5);
        CHECK(Generator::for_piece(empty, e5).size() == 13);
        empty.remove_piece(e5);

        empty.add_piece(Piece(WHITE, BISHOP), Coord::from_fen("a1"));
        CHECK(Generator::for_piece(empty, Coord::from_fen("a1")).size() == 7);
        empty.remove_piece(Coord::from_fen("a1"));

    // Queen
        for (UNumber file = 1; file <= 8; ++file)
            for (UNumber rank = 1; rank <= 8; ++rank)
            {
                auto cell = Coord(file, rank);
                empty.add_piece(Piece(WHITE, ROOK), cell);
                auto rook_size = Generator::for_piece(empty, cell).size();
                empty.remove_piece(cell);
                empty.add_piece(Piece(WHITE, BISHOP), cell);
                auto bishop_size = Generator::for_piece(empty, cell).size();
                empty.remove_piece(cell);
                empty.add_piece(Piece(WHITE, QUEEN), cell);
                auto queen_size = Generator::for_piece(empty, cell).size();
                empty.remove_piece(cell);

                CHECK(queen_size == bishop_size + rook_size);
            }

    // Board
    initial = Board::initial();

    CHECK(Generator::for_board(initial).size() == 20);
}

void testChecker() {
    auto board = Board::initial();
    for (auto square: board.squares) {
        auto coord = square.first;
        CHECK(!Checker::is_attacked(board, coord));
    }

    board.remove_piece(Coord::from_fen("d2"));

    CHECK(Checker::is_attacked(board, Coord::from_fen("d7")));

    board.add_piece(Piece(WHITE, KNIGHT), Coord::from_fen("e5"));
    CHECK(Checker::is_attacked(board, Coord::from_fen("f7")));

    board.add_piece(Piece(WHITE, PAWN), Coord::from_fen("b6"));
    CHECK(Checker::is_attacked(board, Coord::from_fen("b6")));
    CHECK(Checker::is_attacked(board, Coord::from_fen("a7")));
    CHECK(Checker::is_attacked(board, Coord::from_fen("c7")));

    CHECK(Checker::is_check(Board::from_fen("8/2pk4/2Bn1r1p/rPB1R1p1/8/P4P2/1K3P1P/3R4 b - - 2 28")));
    CHECK(Checker::is_check(Board::from_fen("2k5/2p3R1/2B2r2/rPB3pp/2n5/P4P2/1K3P1P/3R4 w - - 2 31 ")));
    CHECK(Checker::is_check(Board::from_fen("2k5/2p3R1/2P5/1rB3pp/2n5/PK3P2/5P1P/3R4 w - - 1 33 ")));
    CHECK(Checker::is_check(Board::from_fen("2k5/2p3R1/2P5/2r3pp/2K5/P4P2/5P1P/3R4 w - - 0 34 ")));
    CHECK(Checker::is_check(Board::from_fen("1k1R4/2p3R1/2P5/2K3pp/8/P4P2/5P1P/8 b - - 2 35 ")));
    CHECK(Checker::is_check(Board::from_fen("3R4/k1R5/2P5/2K3pp/8/P4P2/5P1P/8 b - - 0 36 ")));
    CHECK(Checker::is_check(Board::from_fen("R7/2R5/k1P5/2K3pp/8/P4P2/5P1P/8 b - - 2 37 ")));
}

void testMakeMove() {
    auto board = Board::initial();
    auto moves = Generator::for_board(board);

    for (auto scored_move: moves) {
        auto move = scored_move.second;
        auto data = move.make(board);
        data.restore(board);
        CHECK(board == Board::initial());
    }
}

UNumber test_perft(Board& board, UNumber depth) {
	auto moves = Generator::for_board(board);
	if (depth == 0) {
		return 1;
	}

	UNumber result = 0;
    for (auto scored_move: moves) {
        auto move = scored_move.second;
		auto data = move.make(board);

        auto inlegal = Checker::is_check(board, board.turn_color.inverse());
        if (inlegal) {
            data.restore(board);
            continue;
        }

		auto value = test_perft(board, depth - 1);
		result += value;

		data.restore(board);
	}

	return result;
}

UNumber test_verbose_perft(Board& board, UNumber depth) {
	auto moves = Generator::for_board(board);

	if (depth == 0) {
		return 1;
	}

	UNumber result = 0;
	for (auto move_it = moves.begin(); move_it != moves.end(); ++move_it) {
        auto move = move_it->second;
		auto data = move.make(board);

        auto inlegal = Checker::is_check(board, board.turn_color.inverse());
        if (inlegal) {
            data.restore(board);
            continue;
        }

        auto value = test_perft(board, depth - 1);
        std::cout << move.to_fen() << " - " << value << "\n";
        result += value;

		data.restore(board);
	}

	return result;
}

void testRestoreData() {
    auto b1 = Board::initial();
    CHECK(test_perft(b1, 1) == 20);
	CHECK(test_perft(b1, 2) == 400);
	CHECK(test_perft(b1, 3) == 8902);

    auto b2 = Board::from_fen("r3k2r/p1ppqpb1/bn2pnp1/3PN3/1p2P3/2N2Q1p/PPPBBPPP/R3K2R w KQkq - 0 1");
    CHECK(test_perft(b2, 1) == 48);
	CHECK(test_perft(b2, 2) == 2039);
	auto b2n = Board::from_fen("r3k2r/p1ppqpb1/bnN1pnp1/3P4/1p2P3/2N2Q1p/PPPBBPPP/R3K2R b KQkq - 1 1");
	CHECK(test_perft(b2n, 1) == 41);

    auto b3 = Board::from_fen("8/2p5/3p4/KP5r/1R3p1k/8/4P1P1/8 w - - 0 1");
    CHECK(test_perft(b3, 1) == 14);
    CHECK(test_perft(b3, 2) == 191);
    CHECK(test_perft(b3, 3) == 2812);

    auto b4 = Board::from_fen("r3k2r/Pppp1ppp/1b3nbN/nP6/BBP1P3/q4N2/Pp1P2PP/R2Q1RK1 w kq - 0 1");
    CHECK(test_perft(b4, 1) == 6);
    CHECK(test_perft(b4, 2) == 264);
    CHECK(test_perft(b4, 3) == 9467);

    auto b5 = Board::from_fen("rnbqkb1r/pp1p1ppp/2p5/4P3/2B5/8/PPP1NnPP/RNBQK2R w KQkq - 0 6");
	CHECK(test_perft(b5, 1) == 42);
	CHECK(test_perft(b5, 2) == 1352);
	CHECK(test_perft(b5, 3) == 53392);

    auto b6 = Board::from_fen("r4rk1/1pp1qppp/p1np1n2/2b1p1B1/2B1P1b1/P1NP1N2/1PP1QPPP/R4RK1 w - - 0 10");
    CHECK(test_perft(b6, 1) == 46);
    CHECK(test_perft(b6, 2) == 2079);
}

void testEvaluation() {
    auto board = Board::initial();
    // piece scores
    CHECK(Evaluation::piece(Piece(BLACK, PAWN)) == Evaluation::piece(Piece(WHITE, PAWN)));

    // material score
    CHECK(Evaluation::material(board) == Evaluation::material(board));

    board.remove_piece(Coord::from_fen("e2"));
    CHECK(Evaluation::material(board) == -Evaluation::piece(Piece(WHITE, PAWN)));

    // material sum
    board = Board::initial();
    Score initial_sum = Evaluation::piece(Piece(WHITE, PAWN)) * 16
        + Evaluation::piece(Piece(WHITE, KNIGHT)) * 4
        + Evaluation::piece(Piece(WHITE, BISHOP)) * 4
        + Evaluation::piece(Piece(WHITE, ROOK)) * 4
        + Evaluation::piece(Piece(WHITE, QUEEN)) * 2
        + Evaluation::piece(Piece(WHITE, KING)) * 2;
    CHECK(Evaluation::material_sum(board) == initial_sum);

    // mate score
    Score mate_in_10_for_black = Evaluation::mate() - 10; //mate for black in 10 halfmoves
    Score mate_in_1_for_white = -(Evaluation::mate() - 1); //mate for white in 1 halfmove

    CHECK(Evaluation::is_mate(mate_in_10_for_black));
    CHECK(Evaluation::is_mate(mate_in_1_for_white));
    CHECK(Evaluation::get_mate_half_moves(mate_in_10_for_black) == 10);
    CHECK(Evaluation::get_mate_half_moves(mate_in_1_for_white) == -1);
}

std::map<std::string, void(*)()> tests = {
    {"Pieces", testPiece},
    {"Coords", testCoord},
    {"Colors", testColor},
    {"EnpassantPosibilitys", testEnpassantPosibility},
    {"RookingPosibility", testRookingPosibility},
    {"Boards", testBoard},
    {"Moves", testMove},
    {"Generator", testGenerator},
    {"Checker", testChecker},
    {"Making Moves", testMakeMove},
    {"RestoreData", testRestoreData},
    {"Evaluation", testEvaluation}
};

void runTests() {
    std::vector<std::thread> threads;

    for (auto test: tests) {
        out_mutex.lock();
        std::cout << "Running tests: " << test.first << std::endl;
        out_mutex.unlock();
        threads.push_back(std::thread(test.second));
    }

    for (int i = 0; i < threads.size(); ++i) {
        threads.at(i).join();
    }

    RESULTS;
}

int main() {
	try {
		runTests();
	}
	catch (CustomException e) {
		std::cout << e.what() << std::endl;
	}

    return 0;
}
